'use strict';

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var TransactionStats = require('./TransactionStats');
var FeeRate = require('./FeeRate');

var _require = require('./dataStructures'),
    EstimationResult = _require.EstimationResult,
    MempoolTransaction = _require.MempoolTransaction,
    FeeCalculation = _require.FeeCalculation;

var _require2 = require('./constants'),
    SHORT_BLOCK_PERIODS = _require2.SHORT_BLOCK_PERIODS,
    SHORT_SCALE = _require2.SHORT_SCALE,
    SHORT_DECAY = _require2.SHORT_DECAY,
    MED_BLOCK_PERIODS = _require2.MED_BLOCK_PERIODS,
    MED_SCALE = _require2.MED_SCALE,
    MED_DECAY = _require2.MED_DECAY,
    LONG_BLOCK_PERIODS = _require2.LONG_BLOCK_PERIODS,
    LONG_SCALE = _require2.LONG_SCALE,
    LONG_DECAY = _require2.LONG_DECAY,
    HALF_SUCCESS_PCT = _require2.HALF_SUCCESS_PCT,
    SUCCESS_PCT = _require2.SUCCESS_PCT,
    DOUBLE_SUCCESS_PCT = _require2.DOUBLE_SUCCESS_PCT,
    SUFFICIENT_FEETXS = _require2.SUFFICIENT_FEETXS,
    SUFFICIENT_TXS_SHORT = _require2.SUFFICIENT_TXS_SHORT,
    MIN_BUCKET_FEERATE = _require2.MIN_BUCKET_FEERATE,
    MAX_BUCKET_FEERATE = _require2.MAX_BUCKET_FEERATE,
    FEE_SPACING = _require2.FEE_SPACING,
    INF_FEERATE = _require2.INF_FEERATE,
    OLDEST_ESTIMATE_HISTORY = _require2.OLDEST_ESTIMATE_HISTORY,
    FeeReason = _require2.FeeReason;

var Estimator = function () {
  function Estimator() {
    var bestSeenHeight = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 0;
    var firstRecordedHeight = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;
    var historicalFirst = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 0;
    var historicalBest = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : 0;
    var trackedTxs = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : 0;
    var untrackedTxs = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : 0;

    _classCallCheck(this, Estimator);

    this.buckets = [];
    this.bestSeenHeight = bestSeenHeight;
    this.firstRecordedHeight = firstRecordedHeight;
    this.historicalFirst = historicalFirst;
    this.historicalBest = historicalBest;
    this.trackedTxs = trackedTxs;
    this.untrackedTxs = untrackedTxs;
    for (var bucketBoundary = MIN_BUCKET_FEERATE; bucketBoundary <= MAX_BUCKET_FEERATE; bucketBoundary *= FEE_SPACING) {
      this.buckets.push(bucketBoundary);
    }
    this.buckets.push(INF_FEERATE);

    this.feeStats = new TransactionStats(this.buckets, MED_BLOCK_PERIODS, MED_DECAY, MED_SCALE);
    this.shortStats = new TransactionStats(this.buckets, SHORT_BLOCK_PERIODS, SHORT_DECAY, SHORT_SCALE);
    this.longStats = new TransactionStats(this.buckets, LONG_BLOCK_PERIODS, LONG_DECAY, LONG_SCALE);

    this.mempoolTransactions = new Map();
  }

  _createClass(Estimator, [{
    key: 'removeTx',
    value: function removeTx(hash, inBlock) {
      var transaction = this.mempoolTransactions.get(hash);
      var lastAddedTransactionHash = Array.from(this.mempoolTransactions.keys())[this.mempoolTransactions.size - 1];
      var isLastAdded = lastAddedTransactionHash === hash;
      // todo: Why should it make any sense?
      if (!isLastAdded) {
        this.feeStats.removeTx(transaction.blockHeight, this.bestSeenHeight, transaction.bucketIndex, inBlock);
        this.shortStats.removeTx(transaction.blockHeight, this.bestSeenHeight, transaction.bucketIndex, inBlock);
        this.longStats.removeTx(transaction.blockHeight, this.bestSeenHeight, transaction.bucketIndex, inBlock);
        this.mempoolTransactions.delete(hash);
        return true;
      }
      return false;
    }
  }, {
    key: 'processNewMempoolTransactions',
    value: function processNewMempoolTransactions(rawMempoolTransactions) {
      var _this = this;

      var txids = Object.keys(rawMempoolTransactions);
      txids.forEach(function (hash) {
        var transaction = rawMempoolTransactions[hash];
        transaction.hash = hash;
        // todo: remove true, add fee validation
        _this.addTransactionToMempool(transaction, true);
      });
    }

    /**
     * Adds transaction to mempool.
     * Notice: It is important to process blocks before adding new mempool transactions, because
     * only transactions which entered mempool at same height as last block will be processed.
     * @param transaction mempoolTransactionEntry
     * @param isValidFeeEstimate
     * NOTICE: transaction should have 'hash' property, which is not in raw mempool data
     */

  }, {
    key: 'addTransactionToMempool',
    value: function addTransactionToMempool(transaction, isValidFeeEstimate) {
      var hash = transaction.hash,
          height = transaction.height;

      if (this.mempoolTransactions.has(hash)) {
        return;
      }

      if (height !== this.bestSeenHeight) {
        // Ignore side chains and re-orgs; assuming they are random they don't
        // affect the estimate.  We'll potentially double count transactions in 1-block reorgs.
        // Ignore txs if Estimator is not in sync with chainActive.Tip().
        // It will be synced next time a block is processed.
        return;
      }

      // Only want to be updating estimates when our blockchain is synced,
      // otherwise we'll miscalculate how many blocks its taking to get included.
      if (!isValidFeeEstimate) {
        this.untrackedTxs++;
        return;
      }
      this.trackedTxs++;
      // Fee rates are stored and reported as BTC-per-kb:
      var feeRate = new FeeRate(transaction.fee, transaction.size);

      var bucketIndex = this.feeStats.addTx(height, feeRate.getFeePerK());
      this.shortStats.addTx(height, feeRate.getFeePerK());
      this.longStats.addTx(height, feeRate.getFeePerK());
      this.mempoolTransactions.set(hash, new MempoolTransaction(transaction, bucketIndex));
    }
  }, {
    key: 'processBlockTx',
    value: function processBlockTx(blockHeight, transaction) {
      if (!this.removeTx(transaction.hash, true)) {
        // This transaction wasn't being tracked for fee estimation
        return false;
      }

      // How many blocks did it take for miners to include this transaction?
      // blocksToConfirm is 1-based, so a transaction included in the earliest
      // possible block has confirmation count of 1
      var blocksToConfirm = blockHeight - transaction.height;
      if (blocksToConfirm <= 0) {
        // This can't happen because we don't process transactions from a block with a height
        // lower than our greatest seen height
        return false;
      }

      // Fee rates are stored and reported as BTC-per-kb:
      var feeRate = new FeeRate(transaction.fee, transaction.size);

      this.feeStats.record(blocksToConfirm, feeRate.getFeePerK());
      this.shortStats.record(blocksToConfirm, feeRate.getFeePerK());
      this.longStats.record(blocksToConfirm, feeRate.getFeePerK());
      return true;
    }

    /**
     * @param blockHeight
     * @param txids - transactions included in block
     */

  }, {
    key: 'processBlock',
    value: function processBlock(blockHeight, txids) {
      var _this2 = this;

      if (blockHeight <= this.bestSeenHeight) {
        // Ignore side chains and re-orgs; assuming they are random
        // they don't affect the estimate.
        // And if an attacker can re-org the chain at will, then
        // you've got much bigger problems than "attacker can influence
        // transaction fees."
        return;
      }

      // Must update bestSeenHeight in sync with ClearCurrent so that
      // calls to removeTx (via processBlockTx) correctly calculate age
      // of unconfirmed txs to remove from tracking.
      this.bestSeenHeight = blockHeight;

      // Update unconfirmed circular buffer
      this.feeStats.clearCurrent(blockHeight);
      this.shortStats.clearCurrent(blockHeight);
      this.longStats.clearCurrent(blockHeight);

      // Decay all exponential averages
      this.feeStats.updateMovingAverages();
      this.shortStats.updateMovingAverages();
      this.longStats.updateMovingAverages();

      // Update averages with data points from current block
      txids.forEach(function (txid) {
        if (_this2.mempoolTransactions.has(txid)) {
          _this2.processBlockTx(blockHeight, _this2.mempoolTransactions.get(txid));
        }
      });

      if (this.firstRecordedHeight === 0) {
        this.firstRecordedHeight = this.bestSeenHeight;
      }

      this.trackedTxs = 0;
      this.untrackedTxs = 0;
    }
  }, {
    key: 'blockSpan',
    value: function blockSpan() {
      if (this.firstRecordedHeight === 0) {
        return 0;
      }
      if (this.bestSeenHeight < this.firstRecordedHeight) {
        throw new Error('First recorded height can not me bigger than last seen height');
      }
      return this.bestSeenHeight - this.firstRecordedHeight;
    }
  }, {
    key: 'historicalBlockSpan',
    value: function historicalBlockSpan() {
      if (this.historicalFirst === 0) {
        return 0;
      }
      if (this.historicalBest < this.historicalFirst) {
        throw new Error('First recorded historical height can not me bigger than last seen historical height');
      }

      if (this.bestSeenHeight - this.historicalBest > OLDEST_ESTIMATE_HISTORY) {
        return 0;
      }

      return this.historicalBest - this.historicalFirst;
    }
  }, {
    key: 'maxUsableEstimate',
    value: function maxUsableEstimate() {
      // Block spans are divided by 2 to make sure there are enough potential failing data points for the estimate
      return Math.min(this.longStats.getMaxConfirms(), parseInt(Math.max(this.blockSpan(), this.historicalBlockSpan()) / 2, 10));
    }
  }, {
    key: 'estimateSmartFee',
    value: function estimateSmartFee(confirmationTarget) {
      var isConservative = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;

      var target = confirmationTarget;
      var feeCalculation = new FeeCalculation();
      if (feeCalculation) {
        feeCalculation.desiredTarget = target;
        feeCalculation.returnedTarget = target;
      }

      var median = void 0;
      var halfEst = -1;
      var actualEst = -1;
      var doubleEst = -1;
      var consEst = -1;
      var estimationResult = new EstimationResult();

      // Return failure if trying to analyze a target we're not tracking
      if (target <= 0 || target > this.longStats.getMaxConfirms()) {
        return new FeeRate(0);
      }

      // It's not possible to get reasonable estimates for confTarget of 1
      if (target === 1) {
        target = 2;
      }

      var maxUsableEstimate = this.maxUsableEstimate();
      if (target > maxUsableEstimate) {
        target = maxUsableEstimate;
      }
      if (feeCalculation) feeCalculation.returnedTarget = target;

      if (target <= 1) {
        console.log('Target is to small or we do not have enough data', target);
        return new FeeRate(0);
      }
      /** true is passed to estimateCombined fee for target/2 and target so
       * that we check the max confirms for shorter time horizons as well.
       * This is necessary to preserve monotonically increasing estimates.
       * For non-conservative estimates we do the same thing for 2*target, but
       * for conservative estimates we want to skip these shorter horizons
       * checks for 2*target because we are taking the max over all time
       * horizons so we already have monotonically increasing estimates and
       * the purpose of conservative estimates is not to let short term
       * fluctuations lower our estimates by too much.
       */

      var _estimateCombinedFee = this.estimateCombinedFee(parseInt(target / 2, 10), HALF_SUCCESS_PCT, true);

      var _estimateCombinedFee2 = _slicedToArray(_estimateCombinedFee, 2);

      halfEst = _estimateCombinedFee2[0];
      estimationResult = _estimateCombinedFee2[1];

      if (feeCalculation) {
        feeCalculation.est = estimationResult;
        feeCalculation.reason = FeeReason.HALF_ESTIMATE;
      }
      median = halfEst;

      var _estimateCombinedFee3 = this.estimateCombinedFee(target, SUCCESS_PCT, true);

      var _estimateCombinedFee4 = _slicedToArray(_estimateCombinedFee3, 2);

      actualEst = _estimateCombinedFee4[0];
      estimationResult = _estimateCombinedFee4[1];

      if (actualEst > median) {
        median = actualEst;
        if (feeCalculation) {
          feeCalculation.est = estimationResult;
          feeCalculation.reason = FeeReason.FULL_ESTIMATE;
        }
      }

      var _estimateCombinedFee5 = this.estimateCombinedFee(2 * target, DOUBLE_SUCCESS_PCT, !isConservative);

      var _estimateCombinedFee6 = _slicedToArray(_estimateCombinedFee5, 2);

      doubleEst = _estimateCombinedFee6[0];
      estimationResult = _estimateCombinedFee6[1];

      if (doubleEst > median) {
        median = doubleEst;
        if (feeCalculation) {
          feeCalculation.est = estimationResult;
          feeCalculation.reason = FeeReason.DOUBLE_ESTIMATE;
        }
      }

      if (isConservative || median === -1) {
        var _estimateConservative = this.estimateConservativeFee(2 * target);

        var _estimateConservative2 = _slicedToArray(_estimateConservative, 2);

        consEst = _estimateConservative2[0];
        estimationResult = _estimateConservative2[1];

        if (consEst > median) {
          median = consEst;
          if (feeCalculation) {
            feeCalculation.est = estimationResult;
            feeCalculation.reason = FeeReason.CONSERVATIVE;
          }
        }
      }

      // You can uncoment this line to see some calculation details
      // console.log('fee calculation:', feeCalculation);
      if (median < 0) {
        // error condition
        return new FeeRate(0);
      }

      return FeeRate.fromSatoshisPerK(median);
    }

    /** Return a fee estimate at the required successThreshold from the shortest
     * time horizon which tracks confirmations up to the desired target.  If
     * checkShorterHorizon is requested, also allow short time horizon estimates
     * for a lower target to reduce the given answer */

  }, {
    key: 'estimateCombinedFee',
    value: function estimateCombinedFee(confirmationTarget, successThreshold, checkShorterHorizon) {
      var shortStats = this.shortStats,
          feeStats = this.feeStats,
          longStats = this.longStats;

      var estimate = -1;
      var result = new EstimationResult();

      if (confirmationTarget >= 1 && confirmationTarget <= longStats.getMaxConfirms()) {
        // Find estimate from shortest time horizon possible
        if (confirmationTarget <= shortStats.getMaxConfirms()) {
          var _shortStats$estimateM = shortStats.estimateMedianVal(confirmationTarget, SUFFICIENT_TXS_SHORT, successThreshold, true, this.bestSeenHeight); // short horizon


          var _shortStats$estimateM2 = _slicedToArray(_shortStats$estimateM, 2);

          estimate = _shortStats$estimateM2[0];
          result = _shortStats$estimateM2[1];
        } else if (confirmationTarget <= feeStats.getMaxConfirms()) {
          var _feeStats$estimateMed = feeStats.estimateMedianVal(confirmationTarget, SUFFICIENT_FEETXS, successThreshold, true, this.bestSeenHeight); // medium horizon


          var _feeStats$estimateMed2 = _slicedToArray(_feeStats$estimateMed, 2);

          estimate = _feeStats$estimateMed2[0];
          result = _feeStats$estimateMed2[1];
        } else {
          var _longStats$estimateMe = longStats.estimateMedianVal(confirmationTarget, SUFFICIENT_FEETXS, successThreshold, true, this.bestSeenHeight); // long horizon


          var _longStats$estimateMe2 = _slicedToArray(_longStats$estimateMe, 2);

          estimate = _longStats$estimateMe2[0];
          result = _longStats$estimateMe2[1];
        }

        if (checkShorterHorizon) {
          // If a lower confTarget from a more recent horizon returns a lower answer use it.
          if (confirmationTarget > feeStats.getMaxConfirms()) {
            var _feeStats$estimateMed3 = feeStats.estimateMedianVal(feeStats.getMaxConfirms(), SUFFICIENT_FEETXS, successThreshold, true, this.bestSeenHeight),
                _feeStats$estimateMed4 = _slicedToArray(_feeStats$estimateMed3, 2),
                medMax = _feeStats$estimateMed4[0],
                tempResult = _feeStats$estimateMed4[1];

            if (medMax > 0 && (estimate === -1 || medMax < estimate)) {
              estimate = medMax;
              if (result) {
                result = tempResult;
              }
            }
          }
          if (confirmationTarget > shortStats.getMaxConfirms()) {
            var _shortStats$estimateM3 = shortStats.estimateMedianVal(shortStats.getMaxConfirms(), SUFFICIENT_TXS_SHORT, successThreshold, true, this.bestSeenHeight),
                _shortStats$estimateM4 = _slicedToArray(_shortStats$estimateM3, 2),
                shortMax = _shortStats$estimateM4[0],
                _tempResult = _shortStats$estimateM4[1];

            if (shortMax > 0 && (estimate === -1 || shortMax < estimate)) {
              estimate = shortMax;
              if (result) {
                result = _tempResult;
              }
            }
          }
        }
      }
      return [estimate, result];
    }
  }, {
    key: 'estimateConservativeFee',
    value: function estimateConservativeFee(doubleTarget) {
      var estimate = -1;
      var longEstimate = -1;
      var result = new EstimationResult();
      var longResult = new EstimationResult();
      if (doubleTarget <= this.shortStats.getMaxConfirms()) {
        var _feeStats$estimateMed5 = this.feeStats.estimateMedianVal(doubleTarget, SUFFICIENT_FEETXS, DOUBLE_SUCCESS_PCT, true, this.bestSeenHeight);

        var _feeStats$estimateMed6 = _slicedToArray(_feeStats$estimateMed5, 2);

        estimate = _feeStats$estimateMed6[0];
        result = _feeStats$estimateMed6[1];
      }
      if (doubleTarget <= this.feeStats.getMaxConfirms()) {
        var _longStats$estimateMe3 = this.longStats.estimateMedianVal(doubleTarget, SUFFICIENT_FEETXS, DOUBLE_SUCCESS_PCT, true, this.bestSeenHeight);

        var _longStats$estimateMe4 = _slicedToArray(_longStats$estimateMe3, 2);

        longEstimate = _longStats$estimateMe4[0];
        longResult = _longStats$estimateMe4[1];

        if (longEstimate > estimate) {
          estimate = longEstimate;
          result = longResult;
        }
      }
      return [estimate, result];
    }
  }]);

  return Estimator;
}();

module.exports = Estimator;